/*  
 * Copyright (c) 1999  Machine Vision Holdings Incorporated
 * Author: David Woodhouse <David.Woodhouse@mvhi.com>
 * CoAuthor: Mayk Langer <langer@vsys.de>
 *
 * Template driver used: Copyright (c) 1998  Metro Link Incorporated
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, cpy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Except as contained in this notice, the name of the Metro Link shall not be
 * used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Metro Link.
 *
 */
/* $XFree86: xc/programs/Xserver/hw/xfree86/input/penmount/xf86PM.c,v 1.2 2000/08/11 19:10:46 dawes Exp $ */
/*
      for Penmount USB 
*/
#define _PENMOUNT_C_

#define HALF_DUPLEX_READ "/tmp/penmount_cmd"
#define HALF_DUPLEX_WRITE "/tmp/penmount_dat"
#define CALIBRATION_DATA "/etc/penmount.dat"

#include <asm/types.h>
#include <linux/input.h>

/* keithp - a hack to avoid redefinitions of these in xf86str.h */
#ifdef BUS_PCI
#undef BUS_PCI
#endif
#ifdef BUS_ISA
#undef BUS_ISA
#endif

#include <misc.h>
#include <xf86.h>

#define NEED_XF86_TYPES
#include <xf86_ansic.h>
#include <xf86_OSproc.h>
#include <xf86Xinput.h>
#include <xisb.h>
#include <exevents.h>

#include "xf86PM.h"

#define SYSCALL(call) while(((call) == -1) && (errno == EINTR))
#define ABS(x) ((x) > 0 ? (x) : -(x))

InputDriverRec PENMOUNT = 
{
        1,
        "penmount",
        NULL,
        PenMountPreInit,
        /*PenMountUnInit*/NULL,
        NULL,
        0
};        

#ifdef XFree86LOADER

static XF86ModuleVersionInfo VersionRec =
{
	"penmount",
	MODULEVENDORSTRING,
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	1, 0, 3,
	ABI_CLASS_XINPUT,
	ABI_XINPUT_VERSION,
	MOD_CLASS_XINPUT,
	{0, 0, 0, 0}				/* signature, to be patched into the file by
								 * a tool */
};


static const char *reqSymbols[] = {
	"AddEnabledDevice",
	"ErrorF",
	"InitButtonClassDeviceStruct",
	"InitProximityClassDeviceStruct",
	"InitValuatorAxisStruct",
	"InitValuatorClassDeviceStruct",
	"InitPtrFeedbackClassDeviceStruct",
	"RemoveEnabledDevice",
	"Xcalloc",
	"Xfree",
	"XisbBlockDuration",
	"XisbFree",
	"XisbNew",
	"XisbRead",
	"XisbTrace",
	"screenInfo",
	"xf86AddInputDriver",
	"xf86AllocateInput",
	"xf86CloseSerial",				/* first one */
	"xf86CollectInputOptions",
	"xf86ErrorFVerb",
	"xf86FindOptionValue",
	"xf86GetMotionEvents",
	"xf86GetVerbosity",
	"xf86MotionHistoryAllocate",
	"xf86NameCmp",
	"xf86OpenSerial",
	"xf86CloseSerial",				/* second one ? */
	"xf86OptionListCreate",
	"xf86OptionListMerge",
	"xf86OptionListReport",
	"xf86PostButtonEvent",
	"xf86PostMotionEvent",
	"xf86PostProximityEvent",
	"xf86ProcessCommonOptions",
	"xf86RemoveLocalDevice",
	"xf86ScaleAxis",
	"xf86SetIntOption",
	"xf86SetStrOption",
	"xf86XInputSetScreen",
	"xf86XInputSetSendCoreEvents",
	NULL
};


static pointer PenMountSetupProc(pointer module, pointer options, int *errmaj, int *errmin)
{
	xf86LoaderReqSymLists(reqSymbols, NULL);
	xf86AddInputDriver(&PENMOUNT, module, 0);
	return (pointer) 1;
}

XF86ModuleData penmountModuleData = { &VersionRec, PenMountSetupProc, TearDownProc };


#endif /* XFree86LOADER */


/* 
 * Be sure to set vmin appropriately for your device's protocol. You want to
 * read a full packet before returning
 */
static const char *default_options[] =
{
	"BaudRate", 	"19200",
	"StopBits", 	"1",
	"DataBits", 	"8",
	"Parity", 	"None",
	"Vmin", 	"3",
	"Vtime", 	"1",
	"FlowControl", 	"None",
	NULL,
};

static int      debug_level = 0;
static unsigned char g_last_packet[5];

#define TRACE(lvl, f) {if ((lvl) <= debug_level) f;}


/*****************************************************************************
 *	Function Definitions
 ****************************************************************************/
#define IOCTL_COMMAND_BUFFER_LEN 10
#define IOCTL_SET_CALIBRATION 0xFF

void IOWorkerThread(InputInfoPtr pInfo)
{
    PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
    int fd1, fd2;
    unsigned char cmdbuf[IOCTL_COMMAND_BUFFER_LEN];
    TRACE(4, ErrorF("%sIOWorkerThread.\n", PM_INFO));
    fd1 = open(HALF_DUPLEX_READ, O_RDONLY);
    if(fd1 != -1)
    {
        if(read(fd1, cmdbuf, IOCTL_COMMAND_BUFFER_LEN) > 0)
        {
            TRACE(4, ErrorF("%s: receive %.2X %.2X %.2x %.2x %.2x",
                    PM_INFO, cmdbuf[0], cmdbuf[1], cmdbuf[2],
                    cmdbuf[3], cmdbuf[4]));
            switch(cmdbuf[0])
            {
	            case IOCTL_SET_CALIBRATION:
	            	priv->pmode = cmdbuf[1];
	            	priv->min_x = cmdbuf[2] << 8 | cmdbuf[3];
            		priv->max_x = cmdbuf[4] << 8 | cmdbuf[5];    
            		priv->min_y = cmdbuf[6] << 8 | cmdbuf[7];
            		priv->max_y = cmdbuf[8] << 8 | cmdbuf[9];            	
            		
            		xf86SetIntOption(pInfo->options, "MinX", priv->min_x);
            		xf86SetIntOption(pInfo->options, "MaxX", priv->max_x);
	            	xf86SetIntOption(pInfo->options, "MinY", priv->min_y);	
            		xf86SetIntOption(pInfo->options, "MaxY", priv->max_y);	
	            	xf86SetIntOption(pInfo->options, "PMode", priv->pmode);
	            break;
	            default:
	            break;
            }
            fd2 = open(HALF_DUPLEX_WRITE, O_WRONLY);
            if(fd2 != -1)
            	write(fd2, g_last_packet, IOCTL_COMMAND_BUFFER_LEN);
        }
	}
}

static void TearDownProc(pointer p)
{
}

static void LoadUserCalibration(PenMountPrivatePtr priv)
{
	FILE *f = xf86fopen(CALIBRATION_DATA, "r");
	if(!f) return;
	xf86fscanf(f, "%d", &priv->pmode);
	xf86fscanf(f, "%d", &priv->min_x);
	xf86fscanf(f, "%d", &priv->max_x);
	xf86fscanf(f, "%d", &priv->min_y);
	xf86fscanf(f, "%d", &priv->max_y);
	xf86fclose(f);	
}

static InputInfoPtr PenMountPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
{              
	InputInfoPtr pInfo;
   	PenMountPrivatePtr priv = xcalloc (1, sizeof (PenMountPrivateRec));
	char *s;

	if (!priv)
		return NULL;

	if (!(pInfo = xf86AllocateInput(drv, 0))) 
	{
		xfree(priv);
		return NULL;
	}
  
	priv->pmode = PENMOUNT_DEF_PMODE;
	priv->min_x = PENMOUNT_DEF_MIN_X;
	priv->max_x = PENMOUNT_DEF_MAX_X;
	priv->min_y = PENMOUNT_DEF_MIN_Y;
	priv->max_y = PENMOUNT_DEF_MAX_Y;
	priv->screen_num = 0;
	priv->screen_width = -1;
	priv->screen_height = -1;
	priv->lex_mode = PenMount_byte0;
	priv->swap_xy = 0;
	priv->button_down = FALSE;
	priv->button_number = 2;
	priv->proximity = FALSE;
	priv->chip = CHIP_UNKNOWN;
	priv->pen_down = 0;

	priv->CalibrationMode = 0;

	pInfo->type_name = XI_TOUCHSCREEN;
	pInfo->device_control = DeviceControl;
	pInfo->read_input = ReadInput;
	pInfo->control_proc = ControlProc;
	pInfo->close_proc = CloseProc;
	pInfo->switch_mode = SwitchMode;
	pInfo->conversion_proc = ConvertProc;
	pInfo->dev = NULL;
	pInfo->private = priv;
	pInfo->private_flags = 0;
	pInfo->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
	pInfo->conf_idev = dev;

	xf86CollectInputOptions(pInfo, default_options, NULL);

	xf86OptionListReport( pInfo->options );

	pInfo->fd = xf86OpenSerial (pInfo->options);
	if (pInfo->fd == -1)
	{
		ErrorF("%sPenMount driver unable to open device.\n", PM_ERROR);
		goto SetupProc_fail;
	}
	xf86CloseSerial(pInfo->fd);
	/* 
	 * Process the options for your device like this
	 */
	priv->pmode = xf86SetIntOption( pInfo->options, "PMode",PENMOUNT_DEF_PMODE );
	priv->min_x = xf86SetIntOption( pInfo->options, "MinX", PENMOUNT_DEF_MIN_X );
	priv->max_x = xf86SetIntOption( pInfo->options, "MaxX", PENMOUNT_DEF_MAX_X );
	priv->min_y = xf86SetIntOption( pInfo->options, "MinY", PENMOUNT_DEF_MIN_Y );
	priv->max_y = xf86SetIntOption( pInfo->options, "MaxY", PENMOUNT_DEF_MAX_Y );
	LoadUserCalibration(priv);
	priv->screen_num = xf86SetIntOption( pInfo->options, "ScreenNumber", 0 );
	priv->button_number = xf86SetIntOption( pInfo->options, "ButtonNumber", 2 );
	priv->swap_xy = xf86SetIntOption( pInfo->options, "SwapXY", 0 );
	priv->buffer = NULL;
	
	priv->beep = xf86SetIntOption(pInfo->options, "Beep", 0);
	priv->press_vol = xf86SetIntOption(pInfo->options, "PressVol", 100);
	priv->press_pitch = xf86SetIntOption(pInfo->options, "PressPitch", 880);
	priv->press_dur = xf86SetIntOption(pInfo->options, "PressDur", 15) & 0xff;
	priv->rel_vol = xf86SetIntOption(pInfo->options, "ReleaseVol", 0);
	priv->rel_pitch = xf86SetIntOption(pInfo->options, "ReleasePitch", 1200);
	priv->rel_dur = xf86SetIntOption(pInfo->options, "ReleaseDur", 10) & 0xff;
	priv->rclick_pending = 0;
	debug_level = xf86SetIntOption(pInfo->options, "DebugLevel", 0);

	s = xf86FindOptionValue (pInfo->options, "ReportingMode");
	if ((s) && (xf86NameCmp (s, "raw") == 0))
		priv->reporting_mode = TS_Raw;
	else
		priv->reporting_mode = TS_Scaled;

	priv->proximity = FALSE;
	priv->button_down = FALSE;
	priv->lex_mode = PenMount_byte0;
	
	/* this results in an xstrdup that must be freed later */
	pInfo->name = xf86SetStrOption( pInfo->options, "DeviceName", "PenMount");
	xf86ProcessCommonOptions(pInfo, pInfo->options);

	pInfo->flags |= XI86_CONFIGURED;
	return (pInfo);

  SetupProc_fail:
	if ((pInfo) && (pInfo->fd))
		xf86CloseSerial (pInfo->fd);
		
	if ((pInfo) && (pInfo->name))
		xfree (pInfo->name);

	if ((priv) && (priv->buffer))
		XisbFree (priv->buffer);
		
	if (priv)
		xfree (priv);
	/*return (pInfo);*/
	return NULL;
}

static Bool DeviceControl (DeviceIntPtr dev, int mode)
{
	InputInfoPtr pInfo = dev->public.devicePrivate;
	PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
	unsigned char map[] = {0, 1, 2, 3, 4};

	switch (mode)
	{
	case DEVICE_INIT:
		/*
		 * these have to be here instead of in the SetupProc, because when the
		 * SetupProc is run at server startup, screenInfo is not setup yet
		 */
		priv->screen_width = screenInfo.screens[priv->screen_num]->width;
		priv->screen_height = screenInfo.screens[priv->screen_num]->height;
		
		/*
		 * Device reports button press for 1 button.
		 */
		if (InitButtonClassDeviceStruct (dev, 4, map) == FALSE)
		{
			ErrorF("%sUnable to allocate PenMount ButtonClassDeviceStruct.\n", PM_ERROR);
			return !Success;
		}
		
		/*
		 * Device reports motions on 2 axes in absolute coordinates.
		 * Axes min and max values are reported in raw coordinates.
		 */
		if (InitValuatorClassDeviceStruct (dev, 2, xf86GetMotionEvents,
			pInfo->history_size, Absolute) == FALSE)
		{
			ErrorF("%sUnable to allocate PenMount ValuatorClassDeviceStruct.\n", PM_ERROR);
			return !Success;
		}
		else
		{
			InitValuatorAxisStruct (dev, 0, priv->min_x, priv->max_x,
						PENMOUNT_DEF_MAX_X,
						PENMOUNT_DEF_MIN_X /* min_res */ ,
						PENMOUNT_DEF_MAX_X /* max_res */ );
			InitValuatorAxisStruct (dev, 1, priv->min_y, priv->max_y,
						PENMOUNT_DEF_MAX_Y,
						PENMOUNT_DEF_MIN_Y /* min_res */ ,
						PENMOUNT_DEF_MAX_Y /* max_res */ );
		}
		
		if (InitProximityClassDeviceStruct(dev) == FALSE)
		{
			ErrorF("%sunable to allocate PenMount ProximityClassDeviceStruct.\n", PM_ERROR);
			return !Success;
		}
		
		if (InitPtrFeedbackClassDeviceStruct(dev, PenMountPtrCtrl) == FALSE)
		{
			ErrorF("%sunable to allocate PenMount PtrFeedbackClassDeviceStruct\n", PM_ERROR);
			return !Success;
		}
		
		/* 
		 * Allocate the motion events buffer.
		 */
		xf86MotionHistoryAllocate (pInfo);
		return (Success);
		
	case DEVICE_ON:
		pInfo->fd = xf86OpenSerial(pInfo->options);
		if (pInfo->fd == -1)
		{
			xf86Msg(X_WARNING, "%s: cannot open input device\n", pInfo->name);
			return (!Success);
		}
		else 
		{
			priv->buffer = XisbNew(pInfo->fd, PENMOUNT_BUFFER_SIZE);
			if (!priv->buffer) 
			{
				xf86CloseSerial(pInfo->fd);
				pInfo->fd = -1;
				return (!Success);
			}
			
			/*
			xf86FlushInput(pInfo->fd);
			XisbBlockDuration (priv->buffer, 500000);
			*/
		}

	XisbBlockDuration (priv->buffer, -1);
	priv->lex_mode = PenMount_byte0;
	/* xf86FlushInput(pInfo->fd); */
	AddEnabledDevice (pInfo->fd);
	dev->public.on = TRUE;
	return (Success);
		
	case DEVICE_OFF:
	case DEVICE_CLOSE:
		if (pInfo->fd != -1)
		{ 
			RemoveEnabledDevice (pInfo->fd);
			if (priv->buffer)
			{
				XisbFree(priv->buffer);
				priv->buffer = NULL;
			}
			xf86CloseSerial(pInfo->fd);
		}
		dev->public.on = FALSE;
		return (Success);
	default:
		return (BadValue);
	}
}

/* 
 * The ReadInput function will have to be tailored to your device
 */
static void ReadInput (InputInfoPtr pInfo)
{

#define BUFFER_SIZE 256		/* size of reception buffer */
	
	PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
	int x = priv->oldx;
    int y = priv->oldy;        
    int button = priv->old_button;	
	int temp = 0;
	int pmode = 1;
	struct input_event * event;
    char   eventbuf[BUFFER_SIZE];  /* size of reception buffer */
    ssize_t              len;
    int pass=0;

	SYSCALL(len = read(pInfo->fd, eventbuf, BUFFER_SIZE));
	
	if (len <= 0) 
	{
		ErrorF("Error reading penmont usb device : %s\n", strerror(errno));
		return;
	}
	
	for (event=(struct input_event *)eventbuf;
	     event<(struct input_event *)(eventbuf+len); event++) 
	{
		switch (event->type) 
		{
		case EV_ABS:
			TRACE(5, ErrorF("event->type=%d   ", event->type));	
			TRACE(3, ErrorF("event->code=%d  ", event->code));
			switch (event->code) 
			{
			case ABS_X:
				x = event->value;
				pass=0;
				TRACE(3, ErrorF("x=%d\n", x));		         
			break;
			
			case ABS_Y:
				y = event->value;
				pass=1;
				TRACE(3, ErrorF("y=%d\n", y));		         
			break;
			}
		break; /* EV_ABS */	
		
		case EV_KEY:
			TRACE(5, ErrorF("event->type=%d   ", event->type));	
			TRACE(3, ErrorF("event->code=%d  ", event->code));
			switch (event->code) 
			{	        	
			case BTN_LEFT:
				button=event->value;     
				pass=1;
				TRACE(3, ErrorF("pendown=%d\n", event->value));	                    
			break;
			}
		break; /* EV_KEY */
		default:
		continue;
		} /*switch*/

		if( (ABS(x - priv->oldx)<=0)&&
			(ABS(y - priv->oldx)<=0)&&
			(button == priv->old_button)) 
		{
			TRACE(5, ErrorF("filtered\n"));
			continue;   
		}

		priv->oldx=x;
		priv->oldy=y;
		priv->old_button=button;
		
		g_last_packet[0] = button? 0xFF : 0xBF;
		g_last_packet[1] = x / 128;
		g_last_packet[2] = x % 128;
		g_last_packet[3] = y / 128;
		g_last_packet[4] = y % 128;

		TRACE(4, ErrorF("Last Packet: %x %x %x %x %x\n", g_last_packet[0], g_last_packet[1], g_last_packet[2], g_last_packet[3], g_last_packet[4]));

		if(pass==0)      /*just cord X , process next time with y*/
			continue;   

		/* 
		 * Scaling
		 */
		if (priv->reporting_mode == TS_Scaled)
		{		              	
			pmode = priv->pmode;
		
			if(pmode==5 || pmode==6 || pmode==7 || pmode==8 )        
			{
				temp=x;
				x=y;
				y=temp;
			}	
			
			if(pmode==1 || pmode==2 || pmode==5 || pmode==6 )        
			{        	
				if(x<priv->min_x) x=priv->min_x;
				if(x>priv->max_x) x=priv->max_x;		   	
				x = ((x- priv->min_x)*priv->screen_width)/(priv->max_x - priv->min_x);
			}	
			else
			{
				if(x<priv->min_x) x=priv->min_x;
				if(x>priv->max_x) x=priv->max_x;		   	
				x = ((priv->max_x-x)*priv->screen_width)/(priv->max_x - priv->min_x); 
			}	
		
			if(pmode==1 || pmode==3 || pmode==5 || pmode==7 )        
			{        	
				if(y<priv->min_y) y=priv->min_y;
				if(y>priv->max_y) y=priv->max_y;		   	
				y = ((y-priv->min_y)*priv->screen_height)/(priv->max_y - priv->min_y);
			}	
			else
			{
				if(y<priv->min_y) y=priv->min_y;
				if(y>priv->max_y) y=priv->max_y;		   	
				y = ((priv->max_y-y)*priv->screen_height)/(priv->max_y - priv->min_y);
			}	    
			TRACE(4, ErrorF("%sAfter scale 2, x=%d, y=%d.\n", PM_INFO, x, y));
		}
		
		xf86XInputSetScreen (pInfo, priv->screen_num, x, y);

		/*
		* Send events.
		*
		* We *must* generate a motion before a button change if pointer
		* location has changed as DIX assumes this. This is why we always
		* emit a motion, regardless of the kind of packet processed.
		*/		
		xf86PostMotionEvent (pInfo->dev, TRUE, 0, 2, x, y);
		
		/*
		* Emit a button press or release.
		*/
		if ((priv->button_down == FALSE) && (button == 1))
		{
			xf86PostButtonEvent (pInfo->dev, TRUE,
				1, 1, 0, 2, x, y);
			priv->button_down = TRUE;
			PenMountBeep(priv, 1);
		}
		if ((priv->button_down == TRUE) && (button == 0))
		{
			xf86PostButtonEvent (pInfo->dev, TRUE,
				1, 0, 0, 2, x, y);
			priv->button_down = FALSE;
			PenMountBeep(priv, 0);
		}

		IOWorkerThread(pInfo);   	   	     	     
	}  /*for*/
}	

/* 
 * The ControlProc function may need to be tailored for your device
 */
static int ControlProc (InputInfoPtr pInfo, xDeviceCtl * control)
{
	xDeviceTSCalibrationCtl *c = (xDeviceTSCalibrationCtl *) control;
	PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);
                
        priv->min_x = c->min_x;
        priv->max_x = c->max_x;
        priv->min_y = c->min_y;
        priv->max_y = c->max_y;
 
	return (Success);
}

/* 
 * the CloseProc should not need to be tailored to your device
 */
static void CloseProc (InputInfoPtr pInfo)
{
}

/* 
 * The SwitchMode function may need to be tailored for your device
 */
static int SwitchMode (ClientPtr client, DeviceIntPtr dev, int mode)
{
	InputInfoPtr pInfo = dev->public.devicePrivate;
        PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);


	if ((mode == TS_Raw) || (mode == TS_Scaled))
        {
                priv->reporting_mode = mode;
                return (Success);
        }
        else if ((mode == SendCoreEvents) || (mode == DontSendCoreEvents))
        {
                xf86XInputSetSendCoreEvents (pInfo, (mode == SendCoreEvents));
                return (Success);
        }
        else
                return (!Success);   
}

/* 
 * The ConvertProc function may need to be tailored for your device.
 * This function converts the device's valuator outputs to x and y coordinates
 * to simulate mouse events.
 */
static Bool ConvertProc (InputInfoPtr pInfo,
			 int first,
			 int num,
			 int v0,
			 int v1,
			 int v2,
			 int v3,
			 int v4,
			 int v5,
			 int *x,
			 int *y)
{
	PenMountPrivatePtr priv = (PenMountPrivatePtr) (pInfo->private);

	if (priv->reporting_mode == TS_Raw)
	{
                *x = xf86ScaleAxis (v0, 0, priv->screen_width, priv->min_x,
                                                        priv->max_x);
                *y = xf86ScaleAxis (v1, 0, priv->screen_height, priv->min_y,
                                                        priv->max_y);
		TRACE(1, ErrorF("%sin ConvertProc TS_RAW. \n", PM_INFO));
        }
        else
        {
                *x = v0;
                *y = v1;
	}
	return (TRUE);
}

static void PenMountPtrCtrl(DeviceIntPtr device, PtrCtrl *ctrl)
{
  /* I have no clue what this does, except that registering it stops the 
     X server segfaulting in ProcGetPointerMapping()
     Ho Hum.
  */
}

static void PenMountBeep(PenMountPrivatePtr priv, int press)
{
	if (priv->beep == 0)
		return;

	/* ring release bell */
	if (press == 0)
		xf86SoundKbdBell(priv->rel_vol, priv->rel_pitch, priv->rel_dur);
	else
		xf86SoundKbdBell(priv->press_vol, priv->press_pitch, priv->press_dur);
}
